'use strict'

var
  _ = require('lodash'),
  RandExp = require('randexp'),
  chance = require('chance').Chance(),
  djson = require('describe-json'),
  Promise = require('bluebird');

var _schemas = [];
var _customTypes = [
  {
    name: 'number',
    customType: function () {
      return chance.natural();
    }
  },
  {
    name: 'boolean',
    customType: function () {
      return chance.bool();
    }
  },
  {
    name: 'null',
    customType: function () {
      return null;
    }
  }
];

var _dreamHelper = {
  chance: chance,
  oneOf: function (collection) {
    return _.sample(collection);
  }
};

var _defaultOutput = {
  Dream: 'Hello World'
};

var _genericSchema = {
  name: 'generic',
  schema: {
    default: String
  }
};

function Dream() {

  this._selectedSchema;
  this._output;
  this._dreamHelper = _dreamHelper;
  this._input;
  this._generateData;

  this.defaultSchema = function defaultSchema(schema) {
    _genericSchema = validateAndReturnSchema(schema);
    return this;
  }.bind(this);

  this.useSchema = function useSchema(schema) {
    var
      schemaToUse,
      dreamInstance;

    schemaToUse = validateAndReturnSchema(schema);
    dreamInstance = new Dream();
    dreamInstance.schema(schemaToUse)
    dreamInstance._selectedSchema = schemaToUse;
    dreamInstance._input = this._input;

    return dreamInstance;
  }.bind(this);

  this.input = function input(input) {
    this._dreamHelper.input = this._input = input;
    return (this);
  }.bind(this);

  this.generateSchema = function generateSchema() {
    var
      describedJson,
      schemaName,
      jsonInput,
      validatedJsonInput,
      guessProperties,
      newSchema,
      args = [];

    Array.prototype.push.apply(args, arguments);

    args.forEach(function (argument) {
      switch (typeof (argument)) {
        case 'string':
          schemaName = argument;
          break;
        case 'object':
          jsonInput = argument;
          break;
        case 'boolean':
          guessProperties = argument;
          break;
      }
    });

    validatedJsonInput = _.first(_.flatten([jsonInput]));
    describedJson = djson.describe(validatedJsonInput || {});

    newSchema = {
      name: schemaName || 'generic',
      schema: describedJson
    };

    if (guessProperties === true) {
      newSchema.schema = guessCustomTypes(describedJson);
    };
    addOrReplace(_schemas, newSchema);
    return this;
  }.bind(this);

  this.customType = function customType(typeName, customType) {
    var
      newCustomType = {},
      validTypeName;

    validTypeName = typeof (typeName) === 'string' ? typeName : 'generic';

    if (customType.constructor === RegExp) {
      newCustomType = {
        name: validTypeName,
        customType: function () {
          return new RandExp(customType).gen();
        }
      };
    } else if (typeof (customType) === 'function') {
      newCustomType = {
        name: validTypeName,
        customType: customType
      };
    } else {
      newCustomType = {
        name: validTypeName,
        customType: function () {
          return '[Invalid Custom Type]';
        }
      };
    };

    addOrReplace(_customTypes, newCustomType);

    return this;
  }.bind(this);

  this.output = function output(callback) {
    var
      output,
      generateDataAsync;

    if (typeof (callback) === 'function') {
      if (this._generateData || this._generateFromInput) {
        generateDataAsync = Promise.promisify(generateData);
        generateDataAsync()
          .then(function (outputArray) {
            callback(null, outputArray.length === 1 ? outputArray[0] : outputArray);
          }).catch(function (error) {
            callback(error);
          });
        return this;
      } else {
        generateOutput(callback);
        return this;
      }
    } else {

      if (this._generateData) {
        generateData();
      };

      output = this._output || generateOutput();
      this.cleanse();
      return output;
    }

  }.bind(this);

  this.cleanse = function cleanse() {
    this._output = null;
    this._selectedSchema = null;
  }.bind(this);

  this.flushSchemas = function flushSchemas() {
    _schemas = [];
    return this;
  }.bind(this);

  this.schema = function schema(schema) {
    var
      validatedSchema,
      newSchema,
      args = [];

    Array.prototype.push.apply(args, arguments);

    if (args.length > 1) {
      newSchema = {
        name: typeof (args[0]) === 'string' ? args.shift() : 'generic',
        schema: typeof (args[0]) === 'object' ? args.shift() : {}
      };
    } else {
      newSchema = schema;
    }

    validatedSchema = validateAndReturnSchema(newSchema);

    if (validatedSchema.name === 'generic') {
      this._selectedSchema = validatedSchema;
    } else {

      addOrReplace(_schemas, validatedSchema);

      if (_schemas.length === 1) {
        this._selectedSchema = validatedSchema;
      };
    }

    return this;
  }.bind(this);

  this.generate = function generate(amount, generateRandomData) {
    this._generateData = {
      amount: amount,
      generateRandomData: generateRandomData
    };

    return this;
  }.bind(this);
  
  this.generateFromInput = function generate() {
    this._generateFromInput = true;
    return this;
  }.bind(this);

  this.generateRnd = function generateRnd(amount) {
    return this.generate(amount, true);
  };

  var generateData = function generateData(callback) {
    
    var
      generateOutputFromSchemaAsync = Promise.promisify(generateOutputFromSchema),
      inputArray,
      validInputArray,
      outputGenerationFunction,
      generateRandomData,
      iterations = this._generateData && this._generateData.amount ?  this._generateData.amount : 1,
      outputArray = [],
      i = 0;
      
    validInputArray = this._generateFromInput && Array.isArray(this._input);
    outputGenerationFunction  = typeof (callback) === 'function' ? generateOutputFromSchemaAsync : generateOutputFromSchema;
      
      
    if(validInputArray){
        iterations = this._input.length;
    }

    for (; i < iterations; i++) {
      generateRandomData = this._generateData && this._generateData.generateRandomData ?  this._generateData.generateRandomData : false;
      inputArray = validInputArray ? this._input[i] : null;
      outputArray.push(outputGenerationFunction(selectAvailableSchema(), generateRandomData , inputArray));
    };
    

    if (typeof (callback) === 'function') {;
      return Promise.all(outputArray)
      .then(function(result){
        callback(null, result);
      }).catch(function(error){
        callback(error);
      });
    };

    this._output = outputArray.length === 1 ? outputArray[0] : outputArray;


  }.bind(this);

  var addOrReplace = function addOrReplace(collection, item) {
    var
      index;

    index = _.indexOf(collection, _.find(collection, { name: item.name }));
    if (index >= 0) {
      collection.splice(index, 1, item);
    } else {
      collection.push(item);
    };

    return collection;
  };

  var guessCustomTypes = function guessCustomTypes(schemaObject) {
    var
      customTypeExists,
      temporaryList = [];

    _.forIn(schemaObject, function (value, key) {
      if (typeof (value) === 'object') {
        if (Array.isArray(value)) {
          value.forEach(function (item) {
            if (typeof (item) === 'object') {
              temporaryList.push(guessCustomTypes(item));
            } else {
              temporaryList.push(item.toString());
            };
          });
          schemaObject[key] = temporaryList;
        } else {
          schemaObject[key] = guessCustomTypes(value);
        };
      } else {

        customTypeExists = _.find(_customTypes, { name: key.toString() });

        if (typeof (chance[key.toString()]) === 'function' || customTypeExists !== undefined) {
          schemaObject[key] = key.toString();
        };
      };
    });

    return schemaObject;
  };

  var validateAndReturnSchema = function validateAndReturnSchema(schema) {
    if (isValidSchema(schema)) return schema;

    if (typeof (schema) === 'string') {
      var foundSchema = _.findWhere(_schemas, { name: schema });
      return isValidSchema(foundSchema) ? foundSchema : _genericSchema;
    };

    if (typeof (schema) === 'object') {
      return {
        name: 'generic',
        schema: schema
      };
    };

    return _genericSchema;
  }.bind(this);

  var selectAvailableSchema = function selectAvailableSchema() {
    if (this._selectedSchema) {
      return this._selectedSchema;
    };

    if (thereIsSchema() && _schemas.length === 1) {
      return _schemas[0];
    };

    return _genericSchema;
  }.bind(this);

  var generateOutput = function generateOutput(callback) {
    var
      generateOutputFromSchemaAsync = Promise.promisify(generateOutputFromSchema);

    if (typeof (callback) === 'function') {
      return generateOutputFromSchemaAsync(this._selectedSchema || _defaultOutput)
        .then(function (result) {
          callback(null, result);
        }).catch(function (error) {
          callback(error);
        });
    } else {
      return this._selectedSchema ? generateOutputFromSchema(this._selectedSchema) : _defaultOutput;
    };

  }.bind(this);

  var generateOutputFromSchema = function generateOutputFromSchema(schema, generateValues, inputItem, callback) {
    var
      PromisifiedObject,
      valueFromTypeAsync,
      outputObject = {},
      schemaToUse = validateAndReturnSchema(schema);

    if (typeof (callback) === 'function') {
      valueFromTypeAsync = Promise.promisify(getValueFromType);
      PromisifiedObject = _.mapValues(schemaToUse.schema, function (key) {
        return valueFromTypeAsync(key, true, inputItem);
      });
      return Promise.props(PromisifiedObject)
      .then(function(result){
        callback(null, result);
      })
      .catch(function(error){
        callback(error);
      });;
    };
    _.forIn(schemaToUse.schema, function (value, key) {
      outputObject[key] = getValueFromType(value, generateValues, inputItem);
    });

    return outputObject;
  }.bind(this);

  var getValueFromType = function getValueFromType(propertyType, generateValues, inputItem, callback) {
    var valueAsync = Promise.promisify(getValueFromType);
    var
      value,
      temporaryValue,
      temporaryList = [],
      temporaryObject = {},
      customTypeIndex,
      customTypeNeedle,
      context = this,
      isCallback = typeof(callback) === 'function',
      types = {
        'number': Number,
        'string': String,
        'boolean': Boolean,
        'array': Array,
        'object': Object,
        'function': Function,
        'date': Date
      };

    if (propertyType.constructor === RegExp) {
      if (generateValues) {
        if(isCallback){
          callback(null, new RandExp(propertyType).gen());
        }else{
          return new RandExp(propertyType).gen()
        }
      } else {
        if(isCallback){
          callback(null, types[typeof (new RandExp(propertyType).gen())]());
        }else{
          return types[typeof (new RandExp(propertyType).gen())]();
        }
      };
    };

    switch (typeof (propertyType)) {
      case 'string':
        customTypeNeedle = _.find(_customTypes, { name: propertyType });
        customTypeIndex = _.indexOf(_customTypes, customTypeNeedle);

        if (customTypeIndex >= 0) {
          if(isCallback){
            var customTypeAsync = Promise.promisify(customTypeNeedle.customType);
            return customTypeAsync(_.extend(this._dreamHelper, {inputItem: inputItem}))
            .then(function(result){
              callback(null, result);
            }).catch(function(error){
              callback(error);
            });
          }
          
          temporaryValue = customTypeNeedle.customType(_.extend(this._dreamHelper, {inputItem: inputItem}));
        } else {
          if (propertyType === 'array') {
            temporaryValue = [];
          } else {
            temporaryValue = (typeof (chance[propertyType]) === 'function') ? chance[propertyType]() : '[Unknown Custom Type]';
          };
        };

        if (generateValues) {
          value = temporaryValue;
        } else {
          value = types[typeof (temporaryValue !== undefined ? temporaryValue : '')]();
        };

        break;
      case 'function':

        temporaryValue = propertyType();

        if (propertyType === Date) {
          value = new Date(temporaryValue);
        } else {

          if (generateValues) {
            value = isNative(propertyType) ? chance[typeof (temporaryValue)]() : temporaryValue;
          } else {
            value = Array.isArray(temporaryValue) ? types['array']() : types[typeof (temporaryValue)]();
          }

        }

        break;
      case 'object':

        if (Array.isArray(propertyType)) {
          
          if(isCallback){
            return Promise.map(propertyType, function (item) {
              return valueAsync(item, true);
            }).then(function (result) {
              callback(null, result);
            });
          }else{
            propertyType.forEach(function (item) {
              temporaryList.push(getValueFromType.call(context, item, generateValues, callback));
            });
  
            value = temporaryList;
          }
          
        } else {
          if(isCallback){
            var PromisedObject = _.mapValues(propertyType, function (key) {
              return valueAsync(key, true);
            });
            
            return Promise.props(PromisedObject).then(function (result) {
              callback(null, result);
            });
          }else{
            _.forIn(propertyType, function (value, key) {
              temporaryObject[key] = getValueFromType.call(context, value, generateValues, callback);
            });
            
            value = temporaryObject;
          }
        };

        break;
      default:
        value = '[Invalid Property]';
    };

    if(isCallback){
      callback(null, value);
    }else{
      return value;
    }
  }.bind(this);

  var isValidSchema = function isValidSchema(schema) {
    return _.has(schema, 'name') && _.has(schema, 'schema');
  };

  var thereIsSchema = function thereIsSchema() {
    return _schemas.length > 0;
  };

  var isNative = function isNative(func) {
    return !('prototype' in func);
  };

};

module.exports = new Dream();
